using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using Luticate2.Auth.Interfaces;

namespace Luticate2.Auth.Business.ObjectConverter
{
    public class LuObjectConverterDescriptor : ILuObjectConverterDescriptor
    {
        protected enum PropertyType
        {
            Value,
            Object
        }

        protected class LuObjectConverterDescriptorMemberInfo
        {
            public LambdaExpression ValueExpression { get; set; }

            public PropertyType PropertyType { get; set; }
        }

        protected class LuObjectConverterDescriptorMethodInfo
        {
            public LambdaExpression ValueExpression { get; set; }
        }

        protected delegate LuObjectConverterDescriptorMethodInfo DynamicMethodConverter(MethodInfo methodInfo, IDictionary<Type, Type> types);

        private IDictionary<MemberInfo, LuObjectConverterDescriptorMemberInfo> StaticMemberConverters { get; } = new Dictionary<MemberInfo, LuObjectConverterDescriptorMemberInfo>();

        private IDictionary<MethodInfo, LuObjectConverterDescriptorMethodInfo> StaticMethodConverters { get; } = new Dictionary<MethodInfo, LuObjectConverterDescriptorMethodInfo>();

        private IDictionary<MethodInfo, DynamicMethodConverter> DynamicMethodConverters { get; } = new Dictionary<MethodInfo, DynamicMethodConverter>();

        protected void AddMemberConverter(LambdaExpression memberFrom, LambdaExpression valueExpression, PropertyType propertyType)
        {
            // TODO: check memberInfo != null
            var memberInfo = LuExpressionUtils.GetSingleMemberFromExpression(memberFrom);
            StaticMemberConverters.Add(memberInfo, new LuObjectConverterDescriptorMemberInfo
            {
                ValueExpression = valueExpression,
                PropertyType = propertyType
            });
        }

        protected void AddStaticMethodConverter(LambdaExpression methodFrom, LambdaExpression valueExpression)
        {
            // TODO: check methodInfo != null
            var methodInfo = LuExpressionUtils.GetSingleMethodFromExpression(methodFrom);
            StaticMethodConverters.Add(methodInfo, new LuObjectConverterDescriptorMethodInfo
            {
                ValueExpression = valueExpression
            });
        }

        protected void AddDynamicMethodConverter(LambdaExpression methodFrom, DynamicMethodConverter converter)
        {
            // TODO: check methodInfo != null
            var methodInfo = LuExpressionUtils.GetSingleMethodFromExpression(methodFrom);
            var genericMethodInfo = methodInfo.GetGenericMethodDefinition();
            DynamicMethodConverters.Add(genericMethodInfo, converter);
        }

        protected void AddDynamicMethodConverterTemplate(LambdaExpression methodFrom)
        {
            AddDynamicMethodConverter(methodFrom,
                (info, types) =>
                {
                    var genericMethodInfo = info.GetGenericMethodDefinition();
                    var templateMethodInfo = genericMethodInfo.MakeGenericMethod(info.GetGenericArguments().Select(x => types[x]).ToArray());

                    var lambdaParams = new List<ParameterExpression>
                    {
                        Expression.Parameter(templateMethodInfo.DeclaringType) // TODO what if DeclaringType is static class
                    };
                    var methodParams = templateMethodInfo.GetParameters().Select(x => Expression.Parameter(x.ParameterType)).ToList();
                    lambdaParams.AddRange(methodParams);
                    var methodCallExpression = Expression.Call(templateMethodInfo.IsStatic ? null : lambdaParams[0], templateMethodInfo, methodParams);
                    var lambda = Expression.Lambda(methodCallExpression, lambdaParams);

                    return new LuObjectConverterDescriptorMethodInfo
                    {
                        ValueExpression = lambda
                    };
                }
            );
        }

        public LambdaExpression GetMemberValueExpression(MemberInfo memberInfo, IDictionary<Type, Type> types)
        {
            if (StaticMemberConverters.ContainsKey(memberInfo))
            {
                return StaticMemberConverters[memberInfo].ValueExpression;
            }

            return null;
        }

        public LambdaExpression GetMethodValueExpression(MethodInfo methodInfo, IDictionary<Type, Type> types)
        {
            if (StaticMethodConverters.ContainsKey(methodInfo))
            {
                return StaticMethodConverters[methodInfo].ValueExpression;
            }

            var genericMethodInfo = methodInfo.GetGenericMethodDefinition();
            if (DynamicMethodConverters.ContainsKey(genericMethodInfo))
            {
                var result = DynamicMethodConverters[genericMethodInfo](methodInfo, types);
                if (result != null)
                {
                    return result.ValueExpression;
                }
            }

            return null;
        }
    }

    public class LuObjectConverterDescriptor<TTypeFrom, TTypeTo> : LuObjectConverterDescriptor, ILuObjectConverterDescriptor<TTypeFrom, TTypeTo>
    {
        protected void AddMemberConverter(Expression<Func<TTypeFrom, object>> memberFrom,
            Expression<Func<TTypeTo, object>> valueExpression, PropertyType propertyType)
        {
            AddMemberConverter((LambdaExpression) memberFrom, valueExpression, propertyType);
        }

        protected void AddNullMemberConverter(Expression<Func<TTypeFrom, object>> memberFrom)
        {
            AddMemberConverter(memberFrom, to => default(TTypeTo), PropertyType.Value);
        }
    }
}